@@ -4,16 +4,20 @@ An :class:`~.InputProcessor` receives callbacks for the keystrokes parsed from
 the input in the :class:`~prompt_toolkit.inputstream.InputStream` instance.
 
 The `InputProcessor` will according to the implemented keybindings call the
-correct callbacks when new key presses are feed through `feed_key`.
+correct callbacks when new key presses are feed through `feed`.
 """
 from __future__ import unicode_literals
 from prompt_toolkit.buffer import EditReadOnlyBuffer
 from prompt_toolkit.filters.cli import ViNavigationMode
-from prompt_toolkit.keys import Keys
+from prompt_toolkit.keys import Keys, Key
 from prompt_toolkit.utils import Event
 
+from .registry import Registry
+
+from collections import deque
 from six.moves import range
 import weakref
+import six
 
 __all__ = (
     'InputProcessor',
@@ -23,10 +27,16 @@ __all__ = (
 
 class KeyPress(object):
     """
-    :param key: a `Keys` instance.
+    :param key: A `Keys` instance or text (one character).
     :param data: The received string on stdin. (Often vt100 escape codes.)
     """
-    def __init__(self, key, data):
+    def __init__(self, key, data=None):
+        assert isinstance(key, (six.text_type, Key))
+        assert data is None or isinstance(data, six.text_type)
+
+        if data is None:
+            data = key.name if isinstance(key, Key) else key
+
         self.key = key
         self.data = data
 
@@ -48,8 +58,11 @@ class InputProcessor(object):
         p = InputProcessor(registry)
 
         # Send keys into the processor.
-        p.feed_key(KeyPress(Keys.ControlX, '\x18'))
-        p.feed_key(KeyPress(Keys.ControlC, '\x03')
+        p.feed(KeyPress(Keys.ControlX, '\x18'))
+        p.feed(KeyPress(Keys.ControlC, '\x03')
+
+        # Process all the keys in the queue.
+        p.process_keys()
 
         # Now the ControlX-ControlC callback will be called if this sequence is
         # registered in the registry.
@@ -58,13 +71,23 @@ class InputProcessor(object):
     :param cli_ref: weakref to `CommandLineInterface`.
     """
     def __init__(self, registry, cli_ref):
+        assert isinstance(registry, Registry)
+
         self._registry = registry
         self._cli_ref = cli_ref
-        self.reset()
 
         self.beforeKeyPress = Event(self)
         self.afterKeyPress = Event(self)
 
+        # The queue of keys not yet send to our _process generator/state machine.
+        self.input_queue = deque()
+
+        # The key buffer that is matched in the generator state machine.
+        # (This is at at most the amount of keys that make up for one key binding.)
+        self.key_buffer = []
+
+        self.reset()
+
 #        print(' '.join(set(''.join(map(str, kb.keys)) for kb in registry.key_bindings if all(isinstance(X, unicode) for X in kb.keys))))
 
     def reset(self):
@@ -87,18 +110,7 @@ class InputProcessor(object):
         cli = self._cli_ref()
 
         # Try match, with mode flag
-        with_mode = [b for b in self._registry.get_bindings_for_keys(keys) if b.filter(cli)]
-        if with_mode:
-            return with_mode
-
-        # Try match, where the last key is replaced with 'Any', with mode.
-        keys_any = tuple(keys[:-1] + (Keys.Any,))
-
-        with_mode_any = [b for b in self._registry.get_bindings_for_keys(keys_any) if b.filter(cli)]
-        if with_mode_any:
-            return with_mode_any
-
-        return []
+        return [b for b in self._registry.get_bindings_for_keys(keys) if b.filter(cli)]
 
     def _is_prefix_of_longer_match(self, key_presses):
         """
@@ -122,7 +134,7 @@ class InputProcessor(object):
         Coroutine implementing the key match algorithm. Key strokes are sent
         into this generator, and it calls the appropriate handlers.
         """
-        buffer = []
+        buffer = self.key_buffer
         retry = False
 
         while True:
@@ -144,7 +156,7 @@ class InputProcessor(object):
                 # Exact matches found, call handler.
                 if not is_prefix_of_longer_match and matches:
                     self._call_handler(matches[-1], key_sequence=buffer)
-                    buffer = []
+                    del buffer[:]  # Keep reference.
 
                 # No match found.
                 elif not is_prefix_of_longer_match and not matches:
@@ -156,37 +168,63 @@ class InputProcessor(object):
                         matches = self._get_matches(buffer[:i])
                         if matches:
                             self._call_handler(matches[-1], key_sequence=buffer[:i])
-                            buffer = buffer[i:]
+                            del buffer[:i]
                             found = True
+                            break
 
                     if not found:
-                        buffer = buffer[1:]
+                        del buffer[:1]
 
-    def feed_key(self, key_press):
+    def feed(self, key_press):
         """
-        Send a new :class:`KeyPress` into this processor.
+        Add a new :class:`KeyPress` to the input queue.
+        (Don't forget to call `process_keys` in order to process the queue.)
         """
         assert isinstance(key_press, KeyPress)
+        self.input_queue.append(key_press)
 
-        if key_press.key != Keys.CPRResponse:
-            self.beforeKeyPress.fire()
+    def process_keys(self):
+        """
+        Process all the keys in the `input_queue`.
+        (To be called after `feed`.)
 
-        self._process_coroutine.send(key_press)
+        Note: because of the `feed`/`process_keys` separation, it is
+              possible to call `feed` from inside a key binding.
+              This function keeps looping until the queue is empty.
+        """
+        while self.input_queue:
+            key_press = self.input_queue.popleft()
 
-        if key_press.key != Keys.CPRResponse:
-            self.afterKeyPress.fire()
+            if key_press.key != Keys.CPRResponse:
+                self.beforeKeyPress.fire()
+
+            self._process_coroutine.send(key_press)
+
+            if key_press.key != Keys.CPRResponse:
+                self.afterKeyPress.fire()
+
+        # Invalidate user interface.
+        cli = self._cli_ref()
+        if cli:
+            cli.invalidate()
 
     def _call_handler(self, handler, key_sequence=None):
         arg = self.arg
         self.arg = None
 
-        is_repeat = handler == self._previous_handler
+        event = KeyPressEvent(
+            weakref.ref(self), arg=arg, key_sequence=key_sequence,
+            previous_key_sequence=self._previous_key_sequence,
+            is_repeat=(handler == self._previous_handler))
 
+        # Save the state of the current buffer.
+        cli = event.cli  # Can be `None` (In unit-tests only.)
+
+        if handler.save_before(event) and cli:
+            cli.current_buffer.save_to_undo_stack()
+
+        # Call handler.
         try:
-            event = KeyPressEvent(
-                weakref.ref(self), arg=arg, key_sequence=key_sequence,
-                previous_key_sequence=self._previous_key_sequence,
-                is_repeat=is_repeat)
             handler.call(event)
             self._fix_vi_cursor_position(event)
 
@@ -198,11 +236,6 @@ class InputProcessor(object):
         self._previous_key_sequence = key_sequence
         self._previous_handler = handler
 
-        # Invalidate user interface.
-        cli = self._cli_ref()
-        if cli and handler.invalidate_ui(cli):
-            cli.invalidate()
-
     def _fix_vi_cursor_position(self, event):
         """
         After every command, make sure that if we are in Vi navigation mode, we
@@ -277,7 +310,16 @@ class KeyPressEvent(object):
         """
         Repetition argument.
         """
-        return self._arg or 1
+        if self._arg == '-':
+            return -1
+
+        result = int(self._arg or 1)
+
+        # Don't exceed a million.
+        if int(result) >= 1000000:
+            result = 1
+
+        return result
 
     def append_to_arg_count(self, data):
         """
@@ -288,15 +330,12 @@ class KeyPressEvent(object):
         assert data in '-0123456789'
         current = self._arg
 
-        if current is None:
-            if data == '-':
-                data = '-1'
-            result = int(data)
+        if data == '-':
+            assert current is None or current == '-'
+            result = data
+        elif current is None:
+            result = data
         else:
-            result = int("%s%s" % (current, data))
-
-        # Don't exceed a million.
-        if int(result) >= 1000000:
-            result = None
+            result = "%s%s" % (current, data)
 
         self.input_processor.arg = result
